# -*- coding: utf-8 -*-
"""
/***************************************************************************
 MSPA
                                 A ARCGIS plugin
 Morphological Spatial Pattern Analysis
                              -------------------
        begin                : 2022-04-12
        copyright            : (C) 2022 by European Commission JRC
        email                : peter.vogt@ec.europa.eu
        email                : diego.migliavacca@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import arcpy, os, sys, numpy
from time import gmtime, strftime
try:
    from urllib.request import urlopen # Python 3
except ImportError:
    from urllib2 import urlopen # Python 2

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the .pyt file)."""
        self.label = "MSPA Tools"
        self.alias = "MSPA"

        # List of tool classes associated with this toolbox
        self.tools = [MSPA] 

class MSPA(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "MSPA"
        self.VERSION = 2.0
        self.description = "Analysis and description of image objects including morphometric features and connecting pathways - version " + str(self.VERSION)
        self.canRunInBackground = False
        self.plugin_dir = os.path.dirname(__file__)
        self.install_info = arcpy.GetInstallInfo()

    def getParameterInfo(self):
        """Define parameter definitions"""

        # Input Image parameter
        inputImage = arcpy.Parameter(
            displayName="Input image - [(0),1,2] byte",
            name="InputImage",
            datatype="DEFile",
            parameterType="Required",
            direction="Input"
        )
        inputImage.filter.list = ['tif', 'tiff']

        # FGconn parameter
        fgConn = arcpy.Parameter(
            displayName="FGconn",
            name="FGconn",
            datatype="GPLong",
            parameterType="Required",
            direction="Input"
        )
        fgConn.filter.type = "ValueList"
        fgConn.filter.list = [8, 4]

        # EdgeWidth parameter
        edgeWidth = arcpy.Parameter(
            displayName="EdgeWidth",
            name="EdgeWidth",
            datatype="GPLong",
            parameterType="Required",
            direction="Input"
        )
        edgeWidth.filter.type = "ValueList"
        edgeWidth.filter.list = [1, 2, 3, 4, 5]

        # Transition parameter
        Transition = arcpy.Parameter(
            displayName="Transition",
            name="Transition",
            datatype="GPLong",
            parameterType="Required",
            direction="Input"
        )
        Transition.filter.type = "ValueList"
        Transition.filter.list = [1, 0]

        # Intext parameter
        inText = arcpy.Parameter(
            displayName="Intext",
            name="Intext",
            datatype="GPLong",
            parameterType="Required",
            direction="Input"
        )
        inText.filter.type = "ValueList"
        inText.filter.list = [1, 0]

        # Save Path parameter
        savePath = arcpy.Parameter(
            displayName="Select the folder where to save the output file",
            name="SavePath",
            datatype="DEFolder",
            parameterType="Required",
            direction="Input"
        )

        parameters = [inputImage, fgConn, edgeWidth, Transition, inText, savePath]
        return parameters

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed. This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter. This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        try:
            target_url = 'https://ies-ows.jrc.ec.europa.eu/gtb/GTB/MSPA-plugin/version_mspaARCplugin.txt'
            vrs = urlopen(target_url).read().splitlines()[0]
            floatvrs = float(vrs)
            if floatvrs > self.VERSION:
                messages.AddWarningMessage("There is a new version of the plugin available. Visit https://forest.jrc.ec.europa.eu/en/activities/lpa/mspa/ to get your copy of version " + str(floatvrs))
        except:
            pass

        if self.install_info['ProductName'] == 'ArcGISPro':
            x = arcpy.mp.ArcGISProject("current")
            m = x.listMaps("Map")
            if not m:
                messages.AddErrorMessage("Before proceeding, please add a Map to the project")
                sys.exit(0)

        filepath = parameters[0].valueAsText
        spatial_ref = arcpy.Describe(filepath).spatialReference
        if not spatial_ref:
            messages.AddErrorMessage("The image has no projection. Please load a mspa-compliant image")
            sys.exit(0)
        else:
            dataset = arcpy.Raster(filepath)
            nbands = dataset.bandCount
            bandtype = dataset.pixelType
            xsize = dataset.height
            ysize = dataset.width
            if nbands != 1:
                messages.AddErrorMessage("The image is not a single band image. Please load a mspa-compliant image")
                sys.exit(0)
            elif bandtype != "U8":
                messages.AddErrorMessage("The band type is not Byte, it is " + bandtype + ". Please load a mspa-compliant image")
                sys.exit(0)
            elif xsize * ysize > 65000000:
                messages.AddErrorMessage("The image is bigger than 65 MB. Please load a mspa-compliant image")
                sys.exit(0)
            elif xsize < 50 and ysize < 50:
                messages.AddErrorMessage("The x and/or the y size of the image is smaller than 50 pixels. Please load a mspa-compliant image")
                sys.exit(0)
            else:
                arcpy.CalculateStatistics_management(filepath, skip_existing="SKIP_EXISTING")
                maxval = arcpy.GetRasterProperties_management(filepath, "MAXIMUM").getOutput(0)
                intmax = int(maxval)
                if intmax > 2:
                     messages.AddErrorMessage("The max value for the band in the dataset is bigger than 2 (" + maxval + "). Please load a mspa-compliant image")
                     sys.exit(0)
                elif intmax == 2:
                    myarray = arcpy.RasterToNumPyArray(dataset)
                    hist = numpy.histogram(myarray)
                    minval = arcpy.GetRasterProperties_management(filepath, "MINIMUM").getOutput(0)
                    if int(minval) == 0:
                        secondhistval = hist[0][5]
                    else:
                        secondhistval = hist[0][0]
                    if secondhistval == 0:
                        messages.AddErrorMessage("The histogram values for the band in the dataset are not mspa-compliant. Please load a mspa-compliant image")
                        sys.exit(0)
                    else:
                        savepath = parameters[5].valueAsText
                        outfile = savepath + "\\out.tif"
                        try:
                            os.chdir(self.plugin_dir + "\\resources\\gdal")
                            os.system('gdal_translate.exe -strict -co "TILED=NO" -co "PROFILE=BASELINE" -co "COMPRESS=LZW" ' + filepath + ' ' + outfile)
                        except:
                            messages.AddErrorMessage("The file couldn't be saved in the selected directory. Please check if the directory is writable or select another directory")
                            sys.exit(0)

                        if os.path.isfile(outfile):
                            fgconn = parameters[1].valueAsText
                            edgewidth = parameters[2].valueAsText
                            transition = parameters[3].valueAsText
                            intext = parameters[4].valueAsText
                            try:
                                os.chdir(self.plugin_dir + "\\resources")
                                os.system("mspa_win64.exe -i " + outfile + " -graphfg " + fgconn + " -eew " + edgewidth + " -transition " + transition + " -internal " + intext + " -odir " + savepath + "\\")
                            except:
                                messages.AddErrorMessage("Unable to process the image. Please check if the selected directory is writable. Besides, check if the input image is mspa-compliant")
                                sys.exit(0)
                            if os.path.isfile(outfile):
                                if filepath.lower().endswith('.tif'):
                                    inputfile = filepath[:-4].rsplit('\\', 1)[-1]
                                elif filepath.lower().endswith('.tiff'):
                                    inputfile = filepath[:-5].rsplit('\\', 1)[-1]
                                params = "_" + fgconn + "_" + edgewidth + "_" + transition + "_" + intext
                                inputparams = inputfile + params
                                docname = inputparams + ".tif"
                                movedfile = savepath + "\\" + docname

                                try:
                                    xmin = arcpy.GetRasterProperties_management(filepath, "LEFT").getOutput(0)
                                    xmax = arcpy.GetRasterProperties_management(filepath, "RIGHT").getOutput(0)
                                    ymin = arcpy.GetRasterProperties_management(filepath, "BOTTOM").getOutput(0)
                                    ymax = arcpy.GetRasterProperties_management(filepath, "TOP").getOutput(0)
                                    os.chdir(self.plugin_dir + "\\resources\\gdal")
                                    os.system('gdal_translate.exe -strict -a_ullr ' + xmin + ' ' + ymax + ' ' + xmax + ' ' + ymin + ' -mo "TIFFTAG_DOCUMENTNAME=' + docname + '" -mo "TIFFTAG_IMAGEDESCRIPTION=GTB_MSPA, https://forest.jrc.ec.europa.eu/en/activities/lpa/gtb/ ' + fgconn + '_' + edgewidth + '_' + transition + '_' + intext + '" -co "COMPRESS=LZW" ' + outfile + ' ' + movedfile)
                                except:
                                    os.remove(outfile)
                                    messages.AddErrorMessage("Unable to process the image. Please check if the selected directory is writable. Besides, check if the input image is mspa-compliant")
                                    sys.exit(0)

                                if os.path.isfile(movedfile):
                                    arcpy.DefineProjection_management(movedfile, spatial_ref)
                                    if self.install_info['ProductName'] == 'Desktop':
                                        x = arcpy.mapping.MapDocument("CURRENT")
                                        try:
                                            df = arcpy.mapping.ListDataFrames(x)[0]
                                            layer = arcpy.MakeRasterLayer_management(movedfile, inputparams).getOutput(0)
                                            arcpy.mapping.AddLayer(df, layer, "TOP")
                                        except:
                                            os.remove(outfile)
                                            os.remove(movedfile)
                                            messages.AddErrorMessage("Unable to load the layer")
                                            sys.exit(0)
                                    else:
                                        try:
                                            layer = arcpy.MakeRasterLayer_management(movedfile, inputparams).getOutput(0)
                                            m[0].addLayer(layer, "TOP")
                                        except:
                                            os.remove(outfile)
                                            os.remove(movedfile)
                                            messages.AddErrorMessage("Unable to add the layer to the Map")
                                            sys.exit(0)
                                    messages.AddMessage("GeoTiff image processed successfully!")
                                    os.remove(outfile)
                                else:
                                    os.remove(outfile)
                                    messages.AddErrorMessage("Unable to process the image. Please check if the selected directory is writable. Besides, check if the input image is mspa-compliant")
                                    sys.exit(0)
                            else:
                                messages.AddErrorMessage("Unable to process the image. Please check if the selected directory is writable. Besides, check if the input image is mspa-compliant")
                                sys.exit(0)
                        else:
                            messages.AddErrorMessage("The file couldn't be saved in the selected directory. Please select another directory")
                            sys.exit(0)
